syntax = "proto3";
package custom_types;


message MessageWithMap {
    int32 id = 1;  // Unique ID number
    
    map<string, string> map = 2;
}

message MessageWithOptional {
    int32 id = 1;  // Unique ID number
    
    string text = 2;
    
    optional string optional_text = 3;
}

message MessageWithRepeated {
    int32 id = 1;  // Unique ID number
    
    repeated int32 values = 2;
}

