#!/bin/bash

SCRIPT_DIR=$( cd -- "$( dirname -- "${BASH_SOURCE[0]}" )" &> /dev/null && pwd )

${SCRIPT_DIR}/run_test.sh -i /NonReg -t "not (draft or need_update or ScenarioStatus=Draft or ScenarioStatus=NeedUpdate)" "$@"
